// Copyright (C) 2001 Kai Germaschewski
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

// TODO
// switching w/o threads
// common cmsg_put w/mutex
// stop play/tone thread gracefully

#include "driver.h"
#include <capi20.h>
#include <cerrno>

// FIXME

#include <linux/types.h>

struct capi_profile {
	__u16 ncontroller;	/* number of installed controller */
	__u16 nbchannel;	/* number of B-Channels */
	__u32 goptions;		/* global options */
	__u32 support1;		/* B1 protocols support */
	__u32 support2;		/* B2 protocols support */
	__u32 support3;		/* B3 protocols support */
	__u32 reserved[6];	/* reserved */
	__u32 manu[5];		/* manufacturer specific information */
};

CapiConfig::CapiConfig() :
Keydata("/bayonne/capi20")
{
	static KEYDEF defkeys[] = {
		{"buffersize", "128"},
		{NULL, NULL}
	};

	Load(defkeys);
}

CapiDriver::CapiDriver() :
#ifdef	COMMON_OST_NAMESPACE
Driver(), CapiConfig(), Thread(keythreads.priService())
#else
Driver(), CapiConfig(), Thread(NULL, keythreads.priService())
#endif
{
	static SCRKEYWORDS keywords[] = {
		{"join", (scriptmethod_t)&CapiTrunk::scrJoin, &ScriptCommand::chkHasArgs},
		{"wait", (scriptmethod_t)&CapiTrunk::scrWait, &ScriptCommand::chkHasArgs},
		{NULL, NULL, NULL}};

	struct capi_profile cprofile;
	int err;

	running = false;
	appl_id = 0;
	msg_id = 1;

	if (capi20_isinstalled() != CapiNoError) {
		slog(SLOG_ERROR) << "CAPI: capi not installed - " 
				 << strerror(errno) << endl;
		return;
	}

	capi20_get_profile(0, (CAPI_MESSAGE)&cprofile);
	contr_count = cprofile.ncontroller;
	if (!contr_count) {
		slog(SLOG_ERROR) << "CAPI: no controllers" << endl;
		return;
	}
	port_count = 2 * contr_count;

	err = capi20_register(port_count, 7, getBufferSize(), &appl_id);
	if (err != CapiNoError) {
		slog(SLOG_ERROR) << "CAPI: could not register - " 
				 << strerror(errno) << endl;
		return;
	}

	// dsp devices
	ports = new CapiTrunk *[port_count];
	groups = new TrunkGroup *[port_count];
	memset(ports, 0, sizeof(CapiTrunk *) * port_count);
	memset(groups, 0, sizeof(TrunkGroup *) * port_count);

	ScriptCommand::Load(keywords);

	pipe(evbuf);
	int opts = fcntl(evbuf[1], F_GETFL);
	fcntl(evbuf[1], F_SETFL, opts | O_NONBLOCK);

	slog(SLOG_INFO) << "CAPI: driver loaded. " << contr_count 
			<< " controllers; "<< port_count << " trunks"
			<< endl;
}

CapiDriver::~CapiDriver()
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	Stop();

	if(ports)
		delete[] ports;

	if(groups)
		delete[] groups;

	if (appl_id)
		capi20_release(appl_id);

	close(evbuf[0]);
	close(evbuf[1]);
}

int CapiDriver::Start()
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	int count = 0;
	unsigned err;
	_cmsg cmsg;

	if(active) {
		slog(SLOG_ERROR) << "CAPI: driver already started" << endl;
		return 0;
	}

	slog(SLOG_INFO) << "CAPI: driver starting..." << endl;

	active = true;

	for (int contr = 1; contr <= contr_count; contr++) {
		ports[count++] = new CapiTrunk(contr, 0);
		ports[count++] = new CapiTrunk(contr, 1);

		err = LISTEN_REQ (&cmsg, appl_id, msg_id++,
				  contr,   // Controller
				  0,       // InfoMask
				  0x10011, // CIPMask
				  0,       // CIPMask2
				  (unsigned char *)"",  // CallingPartyNumber
				  (unsigned char *)""); // CallingPartySubaddress
		if (err != CapiNoError) {
			slog(SLOG_ERROR) << "CAPI: LISTEN_REQ err = "
					 << err << endl;
		}
		err = capi20_waitformessage(appl_id, NULL);
		if (err != CapiNoError) {
			slog(SLOG_ERROR) << "CAPI: LISTEN_REQ err = "
					 << err << endl;
		}
		err = capi20_get_cmsg(&cmsg, appl_id);
		if (err != CapiNoError) {
			slog(SLOG_ERROR) << "CAPI: LISTEN_CONF err = "
					 << err << endl;
		}
		if (cmsg.Command != CAPI_LISTEN || 
		    cmsg.Subcommand !=  CAPI_CONF ||
		    cmsg.Info != 0) {
			slog(SLOG_ERROR) << "CAPI: LISTEN_CONF Info = "
					 << cmsg.Info << endl;
		}
	}

	if(!running)
		Thread::Start();

	return count;
}

void CapiDriver::Stop()
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	int id;

	if (!active)
		return;

	if (running) {
		slog(SLOG_DEBUG) << "Terminate()" << endl;
		Terminate();
	}
	slog(SLOG_INFO) << "CAPI: driver stopping..." << endl;

	for(id = 0; id < port_count; ++id) {
		if(ports[id])
			delete (ports[id]);
	}
	memset(ports, 0, sizeof(CapiTrunk *) * port_count);
	active = false;

}

int CapiDriver::getTrunkCount()
{
  //	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	return port_count;
}

Trunk *CapiDriver::getTrunkPort(int id)
{
// 	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << " "
// 			 << id << endl;

	if (id < 0 || id >= port_count)
		return NULL;

	if (!ports)
		return NULL;

	return ports[id];
}

void CapiDriver::handleCapiMessage()
{
//	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	unsigned err;
	unsigned contr;
	_cmsg cmsg;

	err = capi20_get_cmsg(&cmsg, appl_id);
	if (err != CapiNoError) {
		slog(SLOG_ERROR) << "CAPI: capi20_get_cmsg err = "
				 << err << endl;
		return;
	}
	contr = cmsg.adr.adrController & 0xff;
	if (contr < 1 || contr > contr_count) {
		slog(SLOG_ERROR) << capi20_cmsg2str(&cmsg) << endl;
		return;
	}
	if (IS_DATA_B3_IND(&cmsg) || 
	    IS_DATA_B3_CONF(&cmsg) ||
	    IS_CONNECT_ACTIVE_IND(&cmsg) ||
	    IS_DISCONNECT_CONF(&cmsg) || 
	    IS_DISCONNECT_IND(&cmsg) || 
	    IS_CONNECT_B3_CONF(&cmsg) ||
	    IS_CONNECT_B3_IND(&cmsg) ||
	    IS_CONNECT_B3_ACTIVE_IND(&cmsg) ||
	    IS_DISCONNECT_B3_IND(&cmsg) ||
	    IS_FACILITY_CONF(&cmsg) ||
	    IS_FACILITY_IND(&cmsg) ||
	    IS_ALERT_CONF(&cmsg)) {
		for (int i = 2 * (contr-1); i <= 2 * (contr-1) + 1; i++) {
			CapiTrunk *trunk = ports[i];
			TrunkEvent event;

			if (trunk->plci != (cmsg.adr.adrPLCI & 0xffff))
				continue;

			trunk->recvCapiMsg(&cmsg);
			return;
		}
		slog(SLOG_ERROR) << "CAPI: PLCI not found " << endl
				 << capi20_cmsg2str(&cmsg) << endl;
		return;
	}
	if (IS_CONNECT_CONF(&cmsg)) {
		for (int i = 2 * (contr-1); i <= 2 * (contr-1) + 1; i++) {
			CapiTrunk *trunk = ports[i];

			if (!trunk->plci) {
				if (trunk->recvCapiMsg(&cmsg))
					return;
			}
		}
		slog(SLOG_ERROR) << "CAPI: PLCI not found " << endl
				 << capi20_cmsg2str(&cmsg) << endl;
		return;
	}
        if (IS_CONNECT_IND(&cmsg)) {
                for (int i = 2 * (contr-1); i <= 2 * (contr-1) + 1; i++) {
                        CapiTrunk *trunk = ports[i];

                        if (!trunk->plci) {
                                if (trunk->recvCapiMsg(&cmsg))
                                        return;
                        }
                }
                capi20_cmsg_answer(&cmsg);
                cmsg.Reject = 2; // reject
                err = capi20_put_cmsg(&cmsg);
                if (err != CapiNoError) {
                        slog(SLOG_ERROR) << "CAPI: capi20_put_cmsg err = "
                                         << err << endl;
                }
                return;
        }
	slog(SLOG_ERROR) << "CAPI: driver not handled " << endl 
			 << capi20_cmsg2str(&cmsg) << endl;
}

void CapiDriver::Run()
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	int cnt = 0;
	timeout_t timer, expires;
	CapiTrunk *trunk;
	TrunkEvent event;
	struct timeval tv;
	fd_set rfds;
	char dummy;
	int retval;
	
	if (!appl_id)
		return;

	setCancel(THREAD_CANCEL_IMMEDIATE);
	for (;;) {
		timer = ~0;
		for (int i = 0; i < port_count; i++) {
			trunk = ports[i];

		retry:
			expires = trunk->getTimer();
			if(expires > 0)
				if(expires < timer) 
					timer = expires;

			if (!expires) {
				debug->DebugService(trunk, "expires");
				event.id = TRUNK_TIMER_EXPIRED;
				trunk->postEvent(&event);
				goto retry;
			}

		}
		if (timer > 10000) {
			tv.tv_sec = 10; // FIXME
			tv.tv_usec = 0;
		} else {
			tv.tv_sec = timer / 1000;
			tv.tv_usec = (timer % 1000) * 1000;
		}

		FD_ZERO(&rfds);
		FD_SET(capi20_fileno(appl_id), &rfds);
		FD_SET(evbuf[0], &rfds);
		retval = select(MAX(capi20_fileno(appl_id), evbuf[0])+1, 
				&rfds, NULL, NULL, &tv);
		if (retval < 0) {
			slog(SLOG_ERROR) << "select returned "
					 << strerror(errno) << endl;
			break;
		}
		if (FD_ISSET(capi20_fileno(appl_id), &rfds))
			handleCapiMessage();
		if (FD_ISSET(evbuf[0], &rfds))
			read(evbuf[0], &dummy, 1);
	}
}

CapiDriver capiivr;
