// Copyright (C) 2001 Kai Germaschewski
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "driver.h"

CapiPlay::CapiPlay(CapiTrunk *trunk)
	: URLAudio(), Service((Trunk *)trunk, keythreads.priAudio())
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	num_frames = 2;
	frame_size = capiivr.getBufferSize();
}

CapiPlay::~CapiPlay()
{
	slog(SLOG_DEBUG) << "CAPI: ~" << __FUNCTION__ << endl;
	
	if (isThread()) 
		slog(SLOG_WARNING) << "CAPI: called from thread ctx" << endl;
	Terminate();
}

void CapiPlay::Initial()
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	CapiTrunk *trk = (CapiTrunk *) trunk;
	char name[32];
	char buffer[32];
	char *fn;
	struct stat ino;
	char *ann;
	struct tm *dt;
	struct tm tbuf;

	trunk->getName(name);
retry:
	fn = getPlayfile();
	if(!fn) {
		if (data->play.mode == PLAY_MODE_ANY)
			Success();
		slog(SLOG_ERROR) << name << ": no file to play" << endl; 
		Failure();
	}
	
	slog(SLOG_DEBUG) << "CAPI: " << fn << endl;
	stat(fn, &ino);
	Open(fn);	
	if(!isOpen()) {
		if(data->play.mode == PLAY_MODE_ANY || 
		   data->play.mode == PLAY_MODE_ONE)
			goto retry;

		slog(SLOG_ERROR) << name << ": " << fn << ": cannot open" << endl;
		Failure();
	}

	if (data->play.mode == PLAY_MODE_TEMP)
		remove(fn);

	dt = localtime_r(&ino.st_ctime, &tbuf);
	sprintf(buffer, "%04d%02d%02d,%02d%02d%02d",
		dt->tm_year + 1900, dt->tm_mon + 1, dt->tm_mday,
		dt->tm_hour, dt->tm_min, dt->tm_sec);
	trunk->setSymbol(SYM_CREATED, buffer);

	ann = getAnnotation();
	if (ann)
		trunk->setSymbol(SYM_ANNOTATION, ann);
	else
		trunk->setSymbol(SYM_ANNOTATION, "");

	if (getEncoding() != ALAW_AUDIO_ENCODING &&
	    getEncoding() != MULAW_AUDIO_ENCODING) {
		slog(SLOG_ERROR) << name << ": unsupported codec required" << endl;
		Failure();
	}
	trk->play_frames = 0;
}

void CapiPlay::Run()
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	char name[33];	
	CapiDriver *drv = (CapiDriver *) driver;
	CapiTrunk *trk = (CapiTrunk *) trunk;
	audioerror_t status = AUDIO_SUCCESS;
	CapiBuffer buf(frame_size);
	unsigned int err;
	_cmsg cmsg;

	trunk->getName(name);

	while (!stopped) {
		if (trk->play_frames >= num_frames) {
			Wait();
			continue;
		}
		status = getSamples(buf.data, buf.len);
		if (status == AUDIO_READ_FAILURE) 
			break;
		buf.bitReverse();
		err = DATA_B3_REQ(&cmsg, drv->appl_id, drv->msg_id++,
				  trk->ncci,
				  buf.data,
				  buf.len,
				  0,
				  0);
		if (err != CapiNoError) {
				slog(SLOG_ERROR) << "CAPI: DATA_REQ err = "
						 << err << endl;
				break;
		}
		++trk->play_frames;
		if (status == AUDIO_READ_INCOMPLETE) 
			break;
	}
	if (status == AUDIO_READ_FAILURE) {
		slog(SLOG_ERROR) << name << ": failed playback" << endl;
		Failure();
	}
	while (trk->play_frames > 0)
		Wait();

	Success();
}

char *CapiPlay::getContinuation()
{
	char *fn;

	if (data->play.mode == PLAY_MODE_ONE || 
            data->play.mode == PLAY_MODE_TEMP)
		return NULL;
retry:
	fn = getPlayfile();

	if(fn && data->play.mode == PLAY_MODE_ANY) {
		slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << fn << endl;
		if(!canAccess(fn))
			goto retry;
	}
	return fn;
}

timeout_t CapiPlay::Stop(void)
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << endl;

	CapiTrunk *trk = (CapiTrunk *) trunk;

	stopped = true;

	return ~0;
	return trk->play_frames * frame_size * 1000/8000;
}

bool CapiTrunk::playHandler(TrunkEvent *event)
{
	slog(SLOG_DEBUG) << "CAPI: " << __FUNCTION__ << " " 
			 << event->id << endl;

	switch(event->id) {
	case TRUNK_ENTER_STATE:
		endTimer();
		if (!flags.offhook) {
			TrunkSignal(TRUNK_SIGNAL_ERROR);
			setSymbol(SYM_ERROR, "play-on-hook");
			handler = &CapiTrunk::stepHandler;
			return true;
		}
		debug->DebugState(this, "play");
		Trunk::setDTMFDetect();
		thread = new CapiPlay(this);
		thread->Start();
		return true;
	case TRUNK_SERVICE_SUCCESS:
		TrunkSignal(TRUNK_SIGNAL_STEP);
		goto stop;
	case TRUNK_SERVICE_FAILURE:
		setSymbol(SYM_ERROR, "play-failed");
		TrunkSignal(TRUNK_SIGNAL_ERROR);
		goto stop;
	case TRUNK_STOP_STATE:
	stop:
		handler = &CapiTrunk::stepHandler;
		return true;
	}
	return false;
}

