// Copyright (C) 2000 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// As a special exception to the GNU General Public License, permission is
// granted for additional uses of the text contained in its release
// of Bayonne as noted here.
//
// This exception is that permission is hereby granted to link Bayonne 
// with the Dialogic runtime libraries to produce a executable image
// without requiring Dialogic's sources to be supplied so long as each
// each source file so linked contains this exclusion.
//
// This exception does not however invalidate any other reasons why
// the resulting executable file might be covered by the GNU General
// public license or invalidate the licensing requirements of any
// other component or library.
//
// This exception applies only to the code released by OST under the
// name Bayonne.  If you copy code from other releases into a copy of
// Bayonne, as the General Public License permits, the exception does not
// apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
//
// If you write modifications of your own to Bayonne, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice, at which
// point the terms of your modification would be covered under the GPL
// as explicitly stated in "COPYING".

#include "driver.h"
#include <cerrno>

#define	SIGEVTCHK	0x1010
#define SIGBITCHK	0x1111

DialogicConfig::DialogicConfig() :
Keydata("/bayonne/dialogic")
{
	static KEYDEF defkeys[] = {
	{"timeslots", "32"},
	{"voice", "16"},
	{"digital", "2"},
	{"routing", "analog"},
	{"analog", "na_an_io"},
	{"protocol", ""},
	{"buffers", "0"},
	{"interval", "65000"},
	{"delay", "10"},
	{"pri", "2"},
	{"span", "e1"},
	{NULL, NULL}};

	if(chkHosting(NULL))
		Load("/hosting/dialogic");
	Load(defkeys);
}

bool DialogicConfig::getRouting(void)
{
	const char *rtflag = getLast("routing");

	if(!stricmp(rtflag, "digital"))
		return false;

	return true;
}

DialogicDriver::DialogicDriver() :
#ifdef	COMMON_OST_NAMESPACE
Driver(), Thread(keythreads.priService() - 1)
#else
Driver(), Thread(NULL, keythreads.priService() - 1)
#endif
{
	static SCRKEYWORDS keywords[] = {
		{"join", (scriptmethod_t)&DialogicTrunk::scrJoin, &ScriptCommand::chkHasArgs},
		{"wait", (scriptmethod_t)&DialogicTrunk::scrWait, &ScriptCommand::chkHasArgs},
		{NULL, NULL, NULL}};

	DialogicTrunk *trk;
	char d4xname[32];
	char dtiname[32];
	int isdev = -1;
	int dticards, voxcards, pricards;
	int dticard = 1, voxcard = 1, digital = 0, pricard = 0, isdncard = 0;
	int card = 0, dti = 0, is = 0;
	int voxcount, vox, dticount;
	int id = 0;
	int ts, tslots, chdev, tsdev;
	int maxslots = getTimeslots();
	interface_t iface;
	char *span;
	CT_DEVINFO ctinfo;

	level = 0;
	status = DialogicTrunk::status;
	memset(status, ' ', sizeof(DialogicTrunk::status));

	running = false;

	voxcards = getVoice();
	dticards = getDigital();
	pricards = getPRICount();
	if(pricards > dticards)
		dticards = pricards;

	isdn_count = pricards;

	if(!voxcards)
	{
		slog(SLOG_ERROR) << "Dialogic driver missing cards; not initialized" << endl;
		maps = NULL;
		ports = NULL;
		groups = NULL;
		port_count = 0;
		port_used = 0;
		return;
	}

	if(dticards * maxslots > voxcards * 4)
		port_count = dticards * maxslots;
	else
		port_count = voxcards * 8;

	devmap = new devices_t[port_count];	
	ports = new DialogicTrunk *[port_count];
	groups = new TrunkGroup *[port_count];
	maps = new DialogicTrunk *[port_count * 5 + 16];
	ismaps = new DialogicISDN *[port_count * 5 + 16];
	memset(ports, 0, sizeof(DialogicTrunk *) * port_count);
	memset(groups, 0, sizeof(TrunkGroup *) * port_count);
	memset(maps, 0, sizeof(DialogicTrunk *) * port_count * 5 + 16);
	memset(ismaps, 0, sizeof(DialogicISDN *) * port_count * 5 + 16);
	memset(devmap, 0, sizeof(devices_t) * port_count);

#ifdef	DIALOGIC_ISDN
	if(isdn_count)
	{
		isdn = new DialogicISDN *[isdn_count];
		memset(isdn, 0, sizeof(DialogicISDN *) * isdn_count);
	}
#endif

#ifndef	SIGNAL
	slog(SLOG_DEBUG) << "dialogic: setting polled mode" << endl;
	mode = SR_POLLMODE;
	if(sr_setparm(SRL_DEVICE, SR_MODEID, &mode) < 0)
	{
		slog(SLOG_ERROR) << "dialogic: unable to set polled mode" << endl;
		return;
	}
#endif

	while(voxcard <= voxcards)
	{
		sprintf(d4xname, "dxxxB%d", voxcard);
		chdev = dx_open(d4xname, 0);
		if(chdev < 0)
		{
			++voxcard;
			continue;
		}
	
		voxcount = ATDV_SUBDEVS(chdev);
		dx_close(chdev);

		slog(SLOG_DEBUG) << "dialogic: vox card " << voxcard << " of " << voxcards << "; ports=" << voxcount << endl;

		for(vox = 1; vox <= voxcount; ++vox) 
		{
			sprintf(d4xname, "dxxxB%dC%d", voxcard, vox);
			chdev = dx_open(d4xname, 0);
			tsdev = 0;
			if(dx_chgfreq(TID_BUSY1, 450, 0, 35, 0))
				slog(SLOG_DEBUG) << ATDV_ERRMSGP(chdev) << endl;
			if(dx_chgdur(TID_BUSY1, 50, 10, 50, 10))
				slog(SLOG_DEBUG) << ATDV_ERRMSGP(chdev) << endl;
			if(dx_chgrepcnt(TID_BUSY1, 2))
				slog(SLOG_DEBUG) << ATDV_ERRMSGP(chdev) << endl;
			dx_chgfreq(TID_BUSY2, 450, 0, 35, 0);
			dx_chgdur(TID_BUSY2, 25, 5, 25, 5);
			dx_chgrepcnt(TID_BUSY2, 3);
			dx_initcallp(chdev);

			if(!dti && vox == 1)
			{
				if(dx_getctinfo(chdev, &ctinfo))
				{
					iface = NT_ANALOG;
					slog(SLOG_ERROR) << "dialogic: " << ATDV_ERRMSGP(chdev) << endl;
				}
				else
					switch(ctinfo.ct_nettype)
					{
					case CT_NTNONE:
						span = (char *)DialogicConfig::getLast("span");
						if(!stricmp(span, "e1"))
						{
							iface = NT_E1;
							dti = 1;
							dticount = 30;
							break;
						}
						if(!stricmp(span, "t1"))
						{
							iface = NT_T1;
							dti = 1;
							dticount = 24;
							break;
						}
					default:
						iface = NT_ANALOG;
						break;
					case CT_NTT1:
						iface = NT_T1;
						dti = 1;
						dticount = 24;
						break;
					case CT_NTE1:
						iface = NT_E1;
						dti = 1;
						dticount = 30;
						break;
					}
			}
			if(!dti)
			{
				sprintf(devmap[id].gcname, ":P_%s:V_%s",
					getAnalog(), d4xname);
				devmap[id].chdev = chdev;
				devmap[id].tsdev = -1;
				devmap[id].isdev = -1;
				devmap[id].iface = iface;
				++id;
				continue;
			}
			while(dti == 1 && dticard <= dticards)
			{
				sprintf(dtiname, "dtiB%d", dticard);
				tsdev = dt_open(dtiname, 0);
				if(tsdev < 0)
				{
					++dticard;
					continue;
				}
				dticount = ATDV_SUBDEVS(tsdev)+1;
				dt_close(tsdev);
				break;
			}

			if(dti == 1)
				isdev = -1;
#ifdef	DIALOGIC_ISDN
			if(dti == 1 && pricard < pricards)
			{
				if(cc_Open(&isdev, dtiname, 0))
					isdev = -1;
			}
			if(isdev > -1 && dti == 1)
			{
				--dticount;
				++pricard;
				cc_Restart(isdev, EV_SYNC);
				isdn[isdncard] = new DialogicISDN(isdev);
				ismaps[isdev] = isdn[isdncard++];
			}
#endif

			if(dti == 1)
			{
				if(isdev < 0)
					slog(SLOG_DEBUG) << "dialogic: dti card " << dticard << "; timeslots=" << dticount << endl;
				else
					slog(SLOG_DEBUG) << "dialogic: pri card " << dticard << "; timeslots =" << dticount << endl;
			}
 
			if(dti)
			{
				if(dticards)
					if(dticard > dticards+1)
				break;

				if(dti > dticount+1)
				{
					dti = 0;
					break;
				}
				sprintf(dtiname, "dtiB%dT%d", dticard, dti);
#ifdef	DIALOGIC_ISDN
				if(isdev > -1)
				{
					if(cc_Open(&tsdev, dtiname, 0))
					{
						++dti;
						continue;
					}
				}
				else
					tsdev = dt_open(dtiname, 0);	
#else
				tsdev = dt_open(dtiname, 0);
#endif
				sprintf(devmap[id].gcname, ":N_%s:P_%s:V_%s",
					dtiname, getProtocol(), d4xname);
				devmap[id].chdev = chdev;
				devmap[id].tsdev = tsdev;
				devmap[id].iface = iface;
				devmap[id].isdev = isdev;
				++id;
				++dti;
			}
		}
		++voxcard;
	}

	uio.u_seek = uio_seek;
	uio.u_read = uio_read;
	uio.u_write = uio_write;

	tpt.tp_type = IO_EOT;
	tpt.tp_termno = DX_LCOFF;
	tpt.tp_length = 3;
	tpt.tp_flags = TF_LCOFF | TF_10MS;

	if(dx_setuio(uio))
	{
		slog(SLOG_CRITICAL) << "dialogic: cannot set uio" << endl;
		id = 0;
	}
	
	ScriptCommand::Load(keywords);
				
	slog(SLOG_INFO) << "Dialogic driver loaded; capacity=" << id << endl;
	port_used = id;
}

DialogicDriver::~DialogicDriver()
{
	int is;

	if(running)
		Terminate();

	Stop();

#ifdef	DIALOGIC_ISDN
	if(isdn_count && isdn)
		for(is = 0; is < isdn_count; ++is)
		{
			if(isdn[is])
				delete isdn[is];
		}
#endif

	if(ports)
		delete[] ports;

	if(isdn)
		delete[] isdn;

	if(maps)
		delete[] maps;

	if(ismaps)
		delete[] ismaps;

	if(groups)
		delete[] groups;

	if(devmap)
		delete[] devmap;
}

int DialogicDriver::Start(void)
{
	int count = 0;
	int port, tsdev, chdev, isdev;

	if(active)
	{
		slog(SLOG_ERROR) << "driver already started" << endl;
		return 0;
	}

	slog(SLOG_DEBUG) << "allocating trunks..." << endl;
	for(port = 0; port < port_used; ++port)
	{
		chdev = devmap[port].chdev;
		tsdev = devmap[port].tsdev;
		isdev = devmap[port].isdev;
		if(!chdev && !tsdev)
			continue;
		++count;
		ports[port] = new DialogicTrunk(port, &devmap[port]);
		maps[chdev] = ports[port];
		if(tsdev > -1)
			maps[tsdev] = ports[port];
	}

	slog(SLOG_DEBUG) << "starting trunks....." << endl;
	for(port = 0; port < port_used; ++port)
		ports[port]->Start();

	slog(SLOG_INFO) << "driver starting..." << endl;

	if(!running)
		Thread::Start();

	active = true;
	return count;
}

void DialogicDriver::Stop(void)
{
	int port;
	if(!active)
		return;

	for(port = 0; port < port_used; ++port)
		ports[port]->Terminate();		

	if(ports)
		memset(ports, 0, sizeof(DialogicTrunk *) * port_count);

	active = false;
	slog(SLOG_INFO) << "driver stopping..." << endl;
}

void DialogicDriver::Run(void)
{
	int evdev;
	DialogicTrunk *trunk;
	TrunkEvent event;
	DX_CST *cst;
	DV_DIGIT *digbuf;
#ifdef	DIALOGIC_ISDN
	CRN crn;
	int statebuf;
#endif

	unsigned short sig;
	short ind;
	int evtype;
	int digcnt, diglen;
	unsigned long tmask;
	char *d;
	long ldata;
	unsigned code;
	unsigned delay = dialogicivr.getDelay();
	void *datap;
	int	len;

	ccxx_sleep(2000);
	running = true;
	slog(SLOG_DEBUG) << "dialogic: service thread started" << endl;

	setCancel(THREAD_CANCEL_IMMEDIATE);
	for(;;)
	{
		HOLD(NULL)
		if(sr_waitevt(EVT_TIMER) == -1)
		{
			RELEASE
			ccxx_sleep(delay);
			continue;
		}
		evtype = sr_getevttype();
#ifdef	DIALOGIC_ISDN
		switch(evtype)
		{
		case CCEV_ANSWERED:
		case CCEV_ACCEPT:
		case CCEV_OFFERED:
		case CCEV_CONNECTED:
		case CCEV_DROPCALL:
		case CCEV_HOLDACK:
		case CCEV_HOLDREJ:
		case CCEV_MOREDIGITS:
		case CCEV_RELEASECALL:
		case CCEV_RELEASECALLFAIL:
		case CCEV_REQANI:
		case CCEV_RETRIEVEACK:
		case CCEV_RETRIEVEREJ:
		case CCEV_RETRIEVECALL:
		case CCEV_SETBILLING:
		case CCEV_RESTART:
		case CCEV_RESTARTFAIL:
		case CCEV_TASKFAIL:
		case CCEV_ALERTING:
		case CCEV_CALLINFO:
		case CCEV_CONGESTION:
		case CCEV_DISCONNECTED:
		case CCEV_DIVERTED:
		case CCEV_FACILITY:
#ifdef	CCEV_FACILITYACK
		case CCEV_FACILITYACK:
		case CCEV_FACILITYREJ:
#endif
		case CCEV_HOLDCALL:
		case CCEV_L2FRAME:
		case CCEV_L2NOBFFR:
		case CCEV_NOTIFY:
		case CCEV_NSI:
		case CCEV_ISDNMSG:
		case CCEV_PROCEEDING:
		case CCEV_PROGRESSING:
		case CCEV_SETUP_ACK:
#ifdef	CCEV_TERM_REGISTGER
		case CCEV_TERM_REGISTER:
#endif
		case CCEV_TRANSFERACK:
		case CCEV_TRANSFERREJ:
		case CCEV_TRANSIT:
#ifdef	CCEV_USERINFO
		case CCEV_USERINFO:
#endif
		case CCEV_SETCHANSTATE:
		case CCEV_D_CHAN_STATUS:
			datap = sr_getevtdatap();
			cc_GetLineDev(&evdev, datap);
			cc_GetCRN(&crn, datap);
			break;
		default:
			evdev = sr_getevtdev();
		}				
#else
		evdev = sr_getevtdev();
#endif
		if(evdev < 0)
		{
			CONTINUE
		}

		trunk = maps[evdev];
		if(!trunk)
		{
			CONTINUE
		}

		memset(&event, 0, sizeof(event));

#ifdef	DIALOGIC_ISDN
		event.parm.crn = crn;
#endif
		
		switch(evtype)
		{
#ifdef	DIALOGIC_ISDN
		case CCEV_RESTART:
			event.id = TRUNK_CALL_RESTART;
			event.parm.ok = true;
			trunk->postEvent(&event);
			break;
		case CCEV_RESTARTFAIL:
			event.id = TRUNK_CALL_RESTART;
			event.parm.ok = false;
			trunk->postEvent(&event);
			break;
		case CCEV_RELEASECALL:
			event.id = TRUNK_CALL_RELEASE;
			event.parm.ok = true;
			trunk->postEvent(&event);
			break;
		case CCEV_RELEASECALLFAIL:
			event.id = TRUNK_CALL_RELEASE;
			event.parm.ok = false;
			trunk->postEvent(&event);
			break;
		case CCEV_DISCONNECTED:
			event.id = TRUNK_STOP_DISCONNECT;
			trunk->putEvent(&event);
			break;
		case CCEV_CONNECTED:
			event.id = TRUNK_TONE_IDLE;
			trunk->putEvent(&event);
			break;
		case CCEV_PROCEEDING:
		case CCEV_ALERTING:
		case CCEV_PROGRESSING:
			cc_CallState(crn, &statebuf);
			switch(statebuf)
			{
				case CCST_DISCONNECTED:
				case CCST_IDLE:
				case CCST_NULL:
					event.id = TRUNK_TIMER_EXPIRED;
					break;
				case CCST_ALERTING:
				case CCST_DIALING:
				case CCST_ACCEPTED:
				case CCST_CONNECTED:
					event.id = TRUNK_LINE_WINK;
					break;
			}
			
                       	trunk->putEvent(&event);
			break;
		case CCEV_OFFERED:
			event.id = TRUNK_CALLER_ID;
			trunk->putEvent(&event);
			event.id = TRUNK_RINGING_ON;
			trunk->putEvent(&event);
			event.id = TRUNK_RINGING_OFF;
			trunk->putEvent(&event);
			break;
		case CCEV_ANSWERED:
			event.id = TRUNK_OFF_HOOK;
			trunk->putEvent(&event);
			break;
		case CCEV_TASKFAIL:
			trunk->putISDNError();
			break;

#endif
		case TDX_CALLP:
			switch(ATDX_CPTERM(trunk->chdev))
			{
			case CR_BUSY:
				event.id = TRUNK_CPA_BUSYTONE;
				break;
			case CR_CEPT:
				event.id = TRUNK_CPA_INTERCEPT;
				break;
			case CR_CNCT:
				event.id = TRUNK_CPA_CONNECT;
				break;
			case CR_FAXTONE:
				event.id = TRUNK_FAX_DETECT;
				break;
			case CR_NOANS:
				event.id = TRUNK_CPA_NOANSWER;
				break;
			case CR_NODIALTONE:
				event.id = TRUNK_CPA_NODIALTONE;
				break;
			case CR_NORB:
				event.id = TRUNK_CPA_NORINGBACK;
				break;
			case CR_STOPD:
				// I don't think this is important
				break;
			}
			if(event.id)
				trunk->putEvent(&event);
			event.id = TRUNK_TONE_IDLE;
			trunk->putEvent(&event);
			break;
		case TDX_WINK:
			event.id = TRUNK_LINE_WINK;
			trunk->putEvent(&event);
			break;
		case TDX_GETDIG:
			digbuf = trunk->digbuf;
			diglen = strlen(digbuf->dg_value);
			digcnt = 0;
			while(digcnt < diglen)
			{
				event.id = TRUNK_DTMF_KEYUP;
				event.parm.dtmf.duration = 40;
				event.parm.dtmf.digit = trunk->getDigit(digbuf->dg_value[digcnt++]);
				trunk->putEvent(&event);
			}
		case TDX_PLAY:
		case TDX_RECORD:
		case TDX_DIAL:			
		case TDX_PLAYTONE:
		case TDX_ERROR:
			tmask = ATDX_TERMMSK(trunk->chdev);
			switch(trunk->interface)
			{
			case NT_ANALOG:
				if(tmask & TM_LCOFF)
				{
					event.id = TRUNK_LINE_WINK; 
					trunk->putEvent(&event);
				}
				break;
			case NT_T1:
				if((ATDT_TSSGBIT(trunk->tsdev) & DTSG_RCVA) == 0)
				{
					if(trunk->Trunk::flags.offhook)
						event.id = TRUNK_CALL_DISCONNECT;
					else
						event.id = TRUNK_CALL_RINGING;
					trunk->putEvent(&event);
				}
				break;
			case NT_E1:
				if((ATDT_TSSGBIT(trunk->tsdev) & DTSG_RCVA) != 0)
				{
					if(trunk->Trunk::flags.offhook)
						event.id = TRUNK_CALL_DISCONNECT;
					else
						event.id = TRUNK_CALL_RINGING;
					trunk->putEvent(&event);
				}
				break;
			}

			if((tmask & TM_IDDTIME) || (tmask & TM_MAXTIME))
			{
				event.id = TRUNK_TIMER_EXPIRED;
				trunk->putEvent(&event);
			}

			if(tmask & TM_TONE)
			{
				event.id = TRUNK_CPA_DIALTONE;
				trunk->putEvent(&event);
			}

			if(tmask & TM_DIGIT)
			{
			}
	
			switch(evtype)
			{
			case TDX_PLAYTONE:
			case TDX_DIAL:
			case TDX_CALLP:
				event.id = TRUNK_TONE_IDLE;
				break;
			case TDX_PLAY:
			case TDX_RECORD:
				event.id = TRUNK_AUDIO_IDLE;
			}
			if(event.id)
				trunk->putEvent(&event);
				
			break;
		case TDX_SETHOOK:
			cst = (DX_CST *)sr_getevtdatap();
			switch(cst->cst_event)
			{
			case DX_ONHOOK:
				event.id = TRUNK_ON_HOOK;
				trunk->putEvent(&event);
				break;
			case DX_OFFHOOK:
				event.id = TRUNK_OFF_HOOK;
				trunk->putEvent(&event);
			}
			break;
		case TDX_CST:
			cst = (DX_CST *)sr_getevtdatap();
			switch(cst->cst_event)
			{
// DE_RNGOFF is hangup in ring - ??
			case DE_RINGS:
				if(cst->cst_data & 0x0001)
				{
					event.id = TRUNK_CALLER_ID;
					trunk->putEvent(&event);
				}
				event.id = TRUNK_RINGING_ON;
				trunk->putEvent(&event);
				event.id = TRUNK_RINGING_OFF;
				trunk->putEvent(&event);
				break;
			case DE_RNGOFF:
				event.id = TRUNK_TIMER_EXPIRED;
				trunk->putEvent(&event);
				break;
			case DE_LCOF:
			case DE_WINK:
				event.id = TRUNK_LINE_WINK;
				trunk->putEvent(&event);
				break;
			case DE_DIGITS:
				event.id = TRUNK_DTMF_KEYUP;
				event.parm.dtmf.duration = 40;
				event.parm.dtmf.digit = trunk->getDigit(cst->cst_data);
				trunk->putEvent(&event);
				break;
			}
			break;
		case SR_TMOUTEVT:
			event.id = TRUNK_TIMER_EXPIRED;
			trunk->putEvent(&event);
			break;
		case DTEV_SIG:
			cst = (DX_CST *)sr_getevtdatap();
			if(!cst)
			{
				slog(SLOG_ERROR) << "DTEV_SIG: no data" << endl;
				break;
			}
			sig = (unsigned short)(cst->cst_event);
			for(ind = 0; ind < 4; ++ind)
			{
				if(!(sig & (SIGEVTCHK << ind)))
					continue;
				switch(sig & (SIGBITCHK << ind))
				{
				case DTMM_AON:
					if(trunk->interface == NT_T1)
						event.id = TRUNK_CALL_RINGING;
					else
						event.id = TRUNK_CALL_DISCONNECT;
					trunk->putEvent(&event);
					break;
				case DTMM_AOFF:
					if(trunk->interface == NT_E1)
						event.id = TRUNK_CALL_RINGING;
					else
						event.id = TRUNK_CALL_DISCONNECT;
					trunk->putEvent(&event);
					break;
				}
			}
			break;
		}
		CONTINUE
	} 
}

DialogicISDN *DialogicDriver::getISDN(int fd)
{
	if(fd < 0 || fd >= port_count * 5 + 16)
		return NULL;

	if(!ismaps)
		return NULL;

	return ismaps[fd];
}

Trunk *DialogicDriver::getTrunkPort(int id)
{
	if(id < 0 || id >= port_count)
		return NULL;

	if(!ports)
		return NULL;

	return (Trunk *)ports[id];
}

void DialogicDriver::Hold(const char *id)
{
	EnterMutex();
#ifdef	DIALOGIC_DEBUG
	slog(SLOG_DEBUG) << "hold " << id << endl;
#endif
	if(!level)
		sr_hold();
	++level;
}

void DialogicDriver::Release(void)
{
	if(level > 0)
		--level;
	if(!level)
		sr_release();
#ifdef	DIALOGIC_DEBUG
	slog(SLOG_DEBUG) << "release" << endl;
#endif
	LeaveMutex();
}

DialogicDriver dialogicivr;

