// Copyright (C) 2000 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "server.h"

class TcpMonitor : public TCPSocket, public Module, public Mutex, public Server
{
private:
	friend class MonSession;

	MonSession *first, *last;

	inline modtype_t getType(void)
		{return MODULE_FIFO;};

	inline char *getName(void)
		{return "*tcpmon";};

	void Stop(void);
	void Run(void);
	bool OnAccept(const InetHostAddress &ia, tpport_t port);
	bool Command(char **argv, ostream *out);
public:
	TcpMonitor();
} tcpmon;

class MonSession : public TCPSession
{
private:
	MonSession *next, *prev;
	
	void Run(void);
	void Final(void);

	MonSession(TCPSocket &server);
	friend class TcpMonitor;
};

TcpMonitor::TcpMonitor() :
TCPSocket(keynetwork.getAddress(), atoi(getenv("TCPMON"))),
Server(keythreads.priService()), Mutex(), Module()
{
	first = last = NULL;
	addCommand();
}

void TcpMonitor::Stop(void)
{
	MonSession *tcp, *next;

	Terminate();
	tcp = first;
	while(tcp)
	{
		next = tcp->next;
		delete tcp;
		tcp = next;
	}
}

void TcpMonitor::Run(void)
{
	MonSession *client;

	for(;;)
	{
		setCancel(THREAD_CANCEL_IMMEDIATE);
		if(isPendingConnection(~0))
		{
			setCancel(THREAD_CANCEL_DEFERRED);
			client = new MonSession(tcpmon);
			client->Start();
		}
	}
}

bool TcpMonitor::OnAccept(const InetHostAddress &ia, tpport_t port)
{
	slog(SLOG_NOTICE) << "tcpmon: accepting connection " << ia.getHostname() << ":" << port << endl;
	return true;
}

MonSession::MonSession(TCPSocket &server) :
#ifdef	COMMON_OST_NAMESPACE
TCPSession(server)
#else
TCPSession(NULL, server)
#endif
{
	tcpmon.EnterMutex();
	next = prev = NULL;

	if(tcpmon.last)
		tcpmon.last->next = this;

	prev = tcpmon.last;	
	if(!tcpmon.first)
		tcpmon.first = this;
	tcpmon.last = this;
	tcpmon.LeaveMutex();
}		

void MonSession::Final(void)
{
	delete this;
}

void MonSession::Run(void)
{
	const char *node = keyserver.getNode();
	char buffer[1024];
	bool rts;
	char *spc, *tok;
	bool login = false;

	setCancel(THREAD_CANCEL_IMMEDIATE);
	*tcp() << "welcome to Bayonne " << getenv("SERVER_VERSION") << endl;
	*this << endl;
	
	*this << "password:";
	flush();
	while(isPending(SOCKET_PENDING_INPUT, 60000) && !login)
	{
		getline(buffer, 1024);
		if(eof())
			break;
		
		spc = strtok_r(buffer, " \t\r\n", &tok);
		if(!stricmp(spc, keyserver.getPassword()))
		{
			login = true;
			break;
		}
		*this << "password:";
		flush();
	}
	if(login)
	{
		*this << node << "> ";	
		flush();
	}
	while(isPending(SOCKET_PENDING_INPUT, 600000) && login)
	{
		getline(buffer, 1024);
		if(eof())
			break;
	
		spc = buffer;
		while(*spc == ' ' || *spc == '\r' || *spc == '\n')
			++spc;

		if(!*spc)
		{
			*this << node << "> ";
			flush();
			continue;
		}

		if(!strnicmp(spc, "bye", 3))
			break;

		setCancel(THREAD_CANCEL_DEFERRED);
		rts = fifo.Command(buffer, this);
		setCancel(THREAD_CANCEL_IMMEDIATE);
		if(rts)
			*this << "END/command" << endl;
		else
			*this << "ERR/command" << endl;
		*this << node << "> ";
		flush();
	}	
	tcpmon.EnterMutex();
	if(tcpmon.first = this)
		tcpmon.first = next;
	if(tcpmon.last = this)
		tcpmon.last = prev;
	if(next)
		next->prev = prev;
	if(prev)
		prev->next = next;
	slog(SLOG_NOTICE) << "tcpmon: client disconnecting..." << endl;
	tcpmon.LeaveMutex();
}
	
bool TcpMonitor::Command(char **argv, ostream *out)
{
	int port;
	Trunk *trunk;
	TrunkGroup *group;
	char status[1024];
	Module *mod;

	if(strnicmp(argv[0], "dis", 3))
		return false;

	if(!argv[1])
		return false;

	if(!stricmp(argv[1], "version"))
	{
		*out << "version " << getenv("SERVER_VERSION") << endl;
		return true;
	}
	if(!stricmp(argv[1], "status"))
	{
		driver->getStatus(status);
		*out << status << endl;
		return true;
	}
	if(!stricmp(argv[1], "modules"))
	{
		mod = getFirst();
		while(mod)
		{
			*out << mod->getName() << endl;
			mod = mod->getNext();
		}
		return true;
	}
	if(!stricmp(argv[1], "driver"))
	{
		*out << "driver: " << plugins.getDriverName() << endl;
		*out << "mixers: " << driver->getMixers() << endl;
		*out << "ports:  " << driver->getTrunkCount() << endl;
		*out << "used:   " << driver->getTrunkUsed() << endl;		
		return true;
	}
	if(!stricmp(argv[1], "calls"))
	{
		for(port = 0; port < driver->getTrunkCount(); ++ port)
		{
			trunk = driver->getTrunkPort(port);
			if(!trunk)
				continue;
			group = driver->getTrunkGroup(port);
			trunk->getName(status);
			if(trunk->isReady())
			{
				*out << group->getName() << ": " << status << " idle" << endl;
				continue;
			}
			*out << group->getName() << ": " << status << " ";
			*out << trunk->getSymbol(SYM_CALLER) << " ";
			*out << trunk->getSymbol(SYM_DIALED) << " ";
			*out << trunk->getSymbol(SYM_LANGUAGE) << endl;
		}
		return true;
	}	
	return false;
}
 
