// Copyright (C) 2000-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "server.h"
#include <ctype.h>

Scheduler::Scheduler() :
#ifdef	COMMON_OST_NAMESPACE
Thread(keythreads.priScheduler(), keythreads.getStack()), ThreadLock()
#else
Thread(NULL, keythreads.priScheduler(), keythreads.getStack()), ThreadLock()
#endif
{
	char path[65];

	interval = 0;
	altschedule = NULL;

	if(canModify(keypaths.getRunfiles()))
		sprintf(path, "%s/bayonne.sched", keypaths.getRunfiles());
	else
		sprintf(path, "%s/.bayonne.sched", getenv("HOME"));

	if(isFile(path))
		rtmp = open(path, O_WRONLY | O_TRUNC);
	else
		rtmp = creat(path, 0664);
	
	chown(path, keyserver.getUid(), keyserver.getGid());
	sched.close();
}

Scheduler::~Scheduler(void)
{
	Terminate();
	close(rtmp);
	sched.close();
}

void Scheduler::altSchedule(const char *sch)
{
	if(sch)
		slog(SLOG_INFO) << "scheduler: " << sch << ": alternate schedule activated" << endl;
	else
		slog(SLOG_INFO) << "scheduler: alternate schedule cleared" << endl;
	altschedule = sch;
	interval = 0;
}

void Scheduler::Initial(void)
{
	char cfgpath[128];
	char *etc = getenv("CONFIG_KEYDATA");
	if(sched.is_open())
		sched.close();

	if(!etc)
		etc = ETC_PREFIX;

	strcpy(cfgpath, keyserver.getPrefix());
	strcat(cfgpath, "/scheduler.conf");
	if(!canAccess(cfgpath))
	{
		strcpy(cfgpath, etc);
		strcat(cfgpath, "bayonne.sched");
	}

	sched.open(cfgpath);
	if(!sched.is_open())
		slog(SLOG_WARNING) << "scheduler: " << cfgpath << ": cannot access" << endl;
	else
	{
		slog(SLOG_NOTICE) << "scheduler: using " << cfgpath << endl;
		interval = Load();
		Update();
	}
}

void Scheduler::Stop(void)
{
	slog(SLOG_DEBUG) << "scheduler: stopped" << endl;
	Terminate();
	sched.close();
	close(rtmp);
}

void Scheduler::Run(void)
{
	Sync *sync;
	time_t now;

	Update();
	for(;;)
	{
		slog(SLOG_DEBUG) << "scheduler starting..." << endl;
		Sleep(60000);
		Session::Clean();
		time(&now);
		sync = Sync::first;
		while(sync)
		{
			if((sync->runtime + 60 * sync->getInterval()) < now)
			{
				time(&sync->runtime);
				if(sync->isScheduled())
				{
					slog(SLOG_INFO) << "scheduler: " << sync->getSyncName() << " updated" << endl;
					sync->Schedule();
				}
			}
			sync = sync->next;
		}
		
		if(!sched)
			Initial();

		if(!sched.is_open() || interval > 0)
			--interval;
		else if(interval < 1)
		{
			interval = Load();
			Update();
		}
	}
}

void Scheduler::Update(void)
{
	schedtmp tmp;

	TrunkGroup *grp = TrunkGroup::first;
	if(rtmp < 0)
		return;

	lseek(rtmp, 0l, SEEK_SET);
	while(grp)
	{
		memset(&tmp, 0, sizeof(tmp));
		strncpy(tmp.grp, grp->getName(), sizeof(tmp.grp));
		grp->getSchedule(tmp.scr);
		write(rtmp, (char *)&tmp, sizeof(tmp));
		grp = grp->next;
	} 
}

int Scheduler::Load(void)
{
	static char *days[7] = {
		"su", "mo", "tu", "we", "th", "fr", "sa"};

	static char *months[12] = {
		"jan", "feb", "mar", "apr", "may", "jun",
		"jul", "aug", "sep", "oct", "nov", "dec"};

	TrunkGroup *group;
	time_t now;
	struct tm *dt;
	char lbuf[256];
	char *name, *day, *start, *cp;
	char *s, *p;
	int hr, min;
	int id, t1, t2, td;
	char *sp = lbuf;
	struct tm tbuf;
	unsigned dayflag;
	int len;
	int dom = 0;
	int mins;

	time(&now);
	dt = localtime_r(&now, &tbuf);

	if(!sched.is_open())
	{
		group = TrunkGroup::first;
		while(group)
		{
			group->setSchedule(NULL);
			group = group->next;
		}
		return keythreads.getInterval();
	}
	
	sched.seekg(0);
	sched.clear();


	group = TrunkGroup::first;
	while(group)
	{
		group->trump = 0;
		group->planned[0] = 0;
		group = group->next;
	}

	for(;;)
	{
		sched.getline(lbuf, 255);
		if(sched.eof())  
			break;
	
		name = lbuf;
		while(*name && isspace(*name))
			++name;

		if(!*name || *name == '#' || *name == ';')
			continue;

		name = strtok_r(name, " \t\n", &sp);
		day = strtok_r(NULL, " \t\n", &sp);
		start = strtok_r(NULL, " \t\n", &sp);
		cp = strtok_r(NULL, " \t\n", &sp);
		if(!cp || !day || !start)
		{
			slog(SLOG_ERROR) << "error in schedule" << endl;
			continue;
		}

		dayflag = 0;
		if(!stricmp(day, "*"))
			dayflag = 10000;
 
		day = strtok_r(day, ".,;", &sp);
		while(day && !dayflag)
		{
			if(altschedule)
			{
				if(!stricmp(day, altschedule))
				{
					dayflag = 50000;
					break;
				}
			}
			len = strlen(day);
			if(len > 3)
			{
				if(day[len - 2] >= '0' && day[len - 2] <= '9')
					dom = atoi(&day[len - 2]);
				else
					dom = atoi(&day[len - 1]);

				if(dom == dt->tm_mday && !strnicmp(day, months[dt->tm_mon], 3))
				{
					dayflag = 30000;
					break;
				}		
			}

			if(!strnicmp(day, days[dt->tm_wday], 2))
				dayflag = 20000;

			if(!dayflag)
				day = strtok_r(NULL, ".,;", &sp);
		}

		if(!dayflag)
			continue;		

		if (!strcmp(start,"*" )) 
			hr = min = 0;
		else 
		{
			hr = atoi(start);
			min = atoi(strchr(start, ':') + 1);	
		}			

		if(hr > dt->tm_hour || (hr == dt->tm_hour && min > dt->tm_min))
			continue;

		dayflag += 100 * hr + min;
		if(!strcmp(name, "*"))
		{
			if ( !( group = TrunkGroup::first ) ) 
				slog( SLOG_WARNING ) << "scheduler: could not find any groups" << endl;
			while(group)
			{
				if(dayflag > group->trump)
				{
					group->trump = dayflag;
					strcpy(group->planned, cp);
				}
				group = group->next;
			}
			continue;
		}

		p = strtok_r(name, ",;", &sp);	
		while(p)
		{
			group = getGroup(p);
			if(!group) 
			{ 
				slog( SLOG_WARNING ) << "scheduler::unknown group '" << p << "' refereced in schedule" << endl;
				p = strtok_r(NULL, ",;", &sp);
				continue;
			} 
			else if(dayflag > group->trump)
			{
				group->trump = dayflag;
				strcpy(group->planned, cp);
			}
			p = strtok_r(NULL, ",;", &sp);
		}
	}

	group = TrunkGroup::first;
	while(group)
	{
		group->setSchedule(group->planned);
		group = group->next;
	}

	sched.clear(); // Probably we've hit an EOF - so we must clear fail()

	mins = keythreads.getInterval();
	td = ((dt->tm_hour * 60) + dt->tm_min) % mins;
	return mins - td;
}
					
